""" Tests for generic and EVLA CAL manifests """

import json
import logging
import shutil
import sys
from pathlib import Path

# pylint: disable=E0401, E0402, R1721, W0621
import pytest

from ingest_envoy.ingestion_manifest import (
    IngestionManifestBuilder,
    find_manifest,
)
from ingest_envoy.manifest_components import (
    IngestionManifestKey,
    InputScienceProduct,
    InputGroup,
    ManifestParameters,
    OutputScienceProduct,
    AncillaryProduct,
    OutputGroup,
    TARFILE_EXT,
    INGESTION_ARTIFACTS_NAME,
    WEBLOG_FILENAME,
)
from ingest_envoy.utilities import (
    ScienceProductType,
    AncillaryProductType,
    Telescope,
)

# pylint: disable=E0401, E1120

# ingest_path is NOT unused! Don't let IJ remove the import!
from .conftest import (
    ingest_path,
    populate_fake_evla_cal_ingest_path,
    WANTED_FILENAMES,
    UNWANTED,
    find_example_manifest,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

FAKE_LOCATOR = "uid://evla/calibration/doo-wah-ditty-ditty-af123"


@pytest.mark.skip("TODO: broken temporarily, pending fix to output group creation")
def test_filters_cal_input_files(ingest_path: Path):
    """
    We'll be getting calibration/image/eb, etc. science products from a directory under
    /lustre/aoc/cluster/pipeline/{CAPO_PROFILE}/workspaces/staging

    Make sure we take -only- the files to be ingested.

    :param ingest_path: our temporary dir
    """

    populate_fake_evla_cal_ingest_path(ingest_path)
    locator = "uid://evla/calibration/twinkle-twinkle-little-quasar"
    manifest, _ = IngestionManifestBuilder(
        telescope=Telescope.EVLA,
        staging_source_dir=ingest_path,
        sp_type=ScienceProductType.EVLA_CAL.value,
        locator=locator,
    ).build()

    # does manifest contain all the data it should, and none of the data it shouldn't?

    assert manifest.locator == locator
    params = manifest.parameters
    assert not params.reingest and not params.ngas_ingest and not params.calibrate

    input_group = manifest.input_group
    assert len(input_group.science_products) == 1

    output_group = manifest.output_group
    assert len(output_group.science_products) == 1
    assert len(output_group.ancillary_products) == 1
    for product in output_group.ancillary_products:
        if product.filename not in WANTED_FILENAMES:
            assert product.filename.startswith(
                INGESTION_ARTIFACTS_NAME
            ) and product.filename.endswith(TARFILE_EXT)
        assert product.filename not in UNWANTED

    sp_out = output_group.science_products[0]
    assert sp_out.type == ScienceProductType.EVLA_CAL

    assert sp_out.filename in WANTED_FILENAMES
    assert sp_out.filename not in UNWANTED

    shutil.rmtree(ingest_path)


@pytest.mark.skip("TODO: broken temporarily, pending fix to output group creation")
def test_writes_expected_output_files(ingest_path: Path):
    """
    Did the manifest builder produce the manifest file, the weblog, and the science product tar?

    :param ingest_path: ingestion location
    :return:
    """
    populate_fake_evla_cal_ingest_path(ingest_path)
    manifest_file, manifest = IngestionManifestBuilder(
        telescope=Telescope.EVLA,
        staging_source_dir=ingest_path,
        locator="uid://evla/calibration/fee-fi-fo-fum-acdf23",
        sp_type=ScienceProductType.EVLA_CAL.value,
    ).build()
    assert manifest_file
    assert manifest

    ingestion_files = [file for file in ingest_path.iterdir()]

    # at a minimum, we expect the manifest, the ingestion artifact, and the science products tar
    assert len(ingestion_files) >= 3
    mf_json = find_manifest(ingest_path)
    assert mf_json
    tars = [file for file in ingestion_files if file.name.endswith(TARFILE_EXT)]
    assert len(tars) >= 2

    shutil.rmtree(ingest_path)


def test_params_json_well_formed():
    """
    Make sure our ManifestParameters makes nice JSON

    :return:
    """
    params = ManifestParameters(
        telescope=Telescope.EVLA,
        reingest=False,
        ngas_ingest=False,
        calibrate=False,
        staging_source_dir=Path("/home/mchammer/evla/parallel-prod"),
    )

    params_json = params.to_json()
    # if we can dump it, it's good
    json.dumps(params_json)


def test_input_sp_well_formed():
    """
    Make sure our InputScienceProduct makes nice JSON

    :return:
    """
    locator = "uid://evla/calibration/vanilla_heath_bar_crunch_1a23e"
    # single science product
    sp_dict = {
        "locator": locator,
    }

    sp_in = InputScienceProduct(locator=locator)
    assert sp_in.to_json() == sp_dict


def test_input_group_well_formed():
    """
    Make sure our InputGroup makes nice JSON

    :return:
    """
    sp1 = InputScienceProduct(locator="uid://evla/execblock/coffee_heath_bar_crunch_7a23f")
    sp2 = InputScienceProduct(locator="uid://evla/execblock/mint_oreo_omg_omg")

    input_group = InputGroup(science_products=[sp1, sp2])
    assert IngestionManifestKey.SCIENCE_PRODUCTS.value in input_group.to_json().keys()

    sps_found = input_group.science_products
    assert len(sps_found) == 2
    assert sp1 in sps_found
    assert sp2 in sps_found


def test_ancillary_product_well_formed():
    """
    The JSON shouldn't contain empty fields

    :return:
    """
    ap1 = AncillaryProduct(type=AncillaryProductType.LOG_TYPE, filename="without_feathers.tar")
    expected = {"type": AncillaryProductType.LOG_TYPE.value, "filename": "without_feathers.tar"}
    actual = ap1.to_json()

    assert actual == expected


def test_output_group_well_formed():
    """
    Make sure our OutputScienceProduct makes nice JSON

    :return:
    """
    osp = OutputScienceProduct(
        type=ScienceProductType.EVLA_CAL, filename="im_a_lil_calibration.tar"
    )
    ap1 = AncillaryProduct(
        type=AncillaryProductType.PIPELINE_ARTIFACTS, filename="without_feathers.tar"
    )
    ap2 = AncillaryProduct(
        type=AncillaryProductType.PIPELINE_ARTIFACTS, filename="with_feathers.tar"
    )

    expected_json = {
        IngestionManifestKey.SCIENCE_PRODUCTS.value: [osp.to_json()],
        IngestionManifestKey.ANCILLARY_PRODUCTS.value: [ap1.to_json(), ap2.to_json()],
    }
    opg = OutputGroup(science_products=[osp], ancillary_products=[ap1, ap2])
    actual_json = opg.to_json()

    assert actual_json == expected_json


@pytest.mark.skip("TODO: broken temporarily, pending fix to output group creation")
def test_evla_cal_manifest_matches_example(ingest_path: Path):
    """
    Given the correct parameters, manifest that matches _16B_069_cal_manifest.json
    should be generated

    :return:
    """

    expected_dir_name = "/lustre/aoc/cluster/pipeline/dsoc-dev/workspaces/staging/cal_test6"
    example = find_example_manifest("_16B_069_cal_manifest")
    with open(example, "r") as infile:
        expected_json = dict(json.load(infile).items())

    # populate ingestion path with fake files for manifest builder to find
    for filename in [
        "16B-069_sb32814386_1_001.57685.66193635417.testdate.caltables.tar",
        WEBLOG_FILENAME,
    ]:
        file = ingest_path / filename
        file.touch()

    builder = IngestionManifestBuilder(
        staging_source_dir=ingest_path,
        telescope=Telescope.EVLA,
        sp_type=ScienceProductType.EVLA_CAL,
        locator="uid://evla/execblock/48ba4c9d-d7c7-4a8f-9803-1115cd52459b",
    )
    _, manifest_file = builder.build()

    with open(manifest_file, "r") as infile:
        actual_json = dict(json.load(infile).items())

    actual_json[IngestionManifestKey.PARAMETERS.value][
        IngestionManifestKey.INGESTION_PATH.value
    ] = expected_dir_name

    assert actual_json == expected_json

    shutil.rmtree(ingest_path)


def populate_ingest_path_for_manifest_evla_cal_example(ingest_path: Path):
    """
    Create fake input files to match EVLA CAL manifest example

    :param ingest_path:
    :return:
    """
    weblog_file = ingest_path / "weblog.tgz"
    weblog_file.touch()
    cal_file = ingest_path / "XYZ-abc+TMN.O00.tar"
    cal_file.touch()
